@@ -18,7 +18,6 @@ import time
 import cgi
 import os
 import shutil
-import webbrowser
 import smtplib
 import shlex
 from Cookie import BaseCookie
@@ -221,6 +220,9 @@ class TestApp(object):
         # @@: Should this be all non-strings?
         if isinstance(params, (list, tuple, dict)):
             params = urllib.urlencode(params)
+        if hasattr(params, 'items'):
+            # Some other multi-dict like format
+            params = urllib.urlencode(params.items())
         if upload_files:
             params = cgi.parse_qsl(params, keep_blank_values=True)
             content_type, params = self.encode_multipart(
@@ -370,7 +372,8 @@ class TestApp(object):
             c = BaseCookie()
             for name, value in self.cookies.items():
                 c[name] = value
-            req.environ['HTTP_COOKIE'] = str(c).split(': ', 1)[1]
+            hc = '; '.join(['='.join([m.key, m.value]) for m in c.values()])
+            req.environ['HTTP_COOKIE'] = hc
         req.environ['paste.testing'] = True
         req.environ['paste.testing_variables'] = {}
         app = lint.middleware(self.app)
@@ -789,6 +792,9 @@ class TestResponse(object):
         """
         if not isinstance(s, (str, unicode)):
             s = str(s)
+        if isinstance(s, unicode):
+            ## FIXME: we don't know that this response uses utf8:
+            s = s.encode('utf8')
         return (self.body.find(s) != -1
                 or self.normal_body.find(s) != -1)
 
@@ -819,7 +825,7 @@ class TestResponse(object):
                     "Body does not contain string %r" % s)
         for no_s in no:
             if no_s in self:
-                print >> sys.stderr, "Actual response (has %r)" % s
+                print >> sys.stderr, "Actual response (has %r)" % no_s
                 print >> sys.stderr, self
                 raise IndexError(
                     "Body contains string %r" % s)
@@ -840,6 +846,7 @@ class TestResponse(object):
         Show this response in a browser window (for debugging purposes,
         when it's hard to read the HTML).
         """
+        import webbrowser
         fn = tempnam_no_warning(None, 'paste-fixture') + '.html'
         f = open(fn, 'wb')
         f.write(self.body)
@@ -1095,6 +1102,8 @@ class Form(object):
             field = self.get(name, index=index)
             submit.append((field.name, field.value_if_submitted()))
         for name, fields in self.fields.items():
+            if name is None:
+                continue
             for field in fields:
                 value = field.value
                 if value is None:
@@ -1221,8 +1230,7 @@ class Checkbox(Field):
     def value__get(self):
         if self.checked:
             if self._value is None:
-                # @@: 'on'?
-                return 'checked'
+                return 'on'
             else:
                 return self._value
         else:
@@ -1518,7 +1526,7 @@ class ProcResult(object):
 
     def assert_no_error(self):
         __tracebackhide__ = True
-        assert self.returncode is 0, (
+        assert self.returncode == 0, (
             "Script returned code: %s" % self.returncode)
 
     def assert_no_stderr(self):
@@ -1689,8 +1697,13 @@ def setup_module(module=None):
 
     Use like::
 
-        from paste.tests.fixture import setup_module
+        from paste.fixture import setup_module
     """
+    # Deprecated June 2008
+    import warnings
+    warnings.warn(
+        'setup_module is deprecated',
+        DeprecationWarning, 2)
     if module is None:
         # The module we were called from must be the module...
         module = sys._getframe().f_back.f_globals['__name__']
