@@ -1,7 +1,6 @@
 # (c) 2005 Ben Bangert
 # This module is part of the Python Paste Project and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
-
 """Registry for handling request-local module globals sanely
 
 Dealing with module globals in a thread-safe way is good if your
@@ -24,7 +23,7 @@ WSGI application everything will work properly.
 
 Example:
 
-.. code-block:: Python
+.. code-block:: python
     
     #yourpackage/__init__.py
     
@@ -64,7 +63,7 @@ thousands of times per request, the performance of the proxy will start to
 become more noticeable. In that circumstance, the problem can be avoided by
 getting at the actual object via the proxy with the ``_current_obj`` function:
 
-.. code-block:: Python
+.. code-block:: python
     
     #sessions.py
     Session = StackedObjectProxy()
@@ -91,7 +90,6 @@ quick way to work around it is documented.
 
 """
 import sys
-import warnings
 import paste.util.threadinglocal as threadinglocal
 
 __all__ = ['StackedObjectProxy', 'RegistryManager', 'StackedObjectRestorer',
@@ -122,6 +120,18 @@ class StackedObjectProxy(object):
         self.__dict__['____local__'] = threadinglocal.local()
         if default is not NoDefault:
             self.__dict__['____default_object__'] = default
+
+    def __dir__(self):
+        """Return a list of the StackedObjectProxy's and proxied
+        object's (if one exists) names.
+        """
+        dir_list = dir(self.__class__) + self.__dict__.keys()
+        try:
+            dir_list.extend(dir(self._current_obj()))
+        except TypeError:
+            pass
+        dir_list.sort()
+        return dir_list
     
     def __getattr__(self, attr):
         return getattr(self._current_obj(), attr)
@@ -140,14 +150,17 @@ class StackedObjectProxy(object):
     
     def __delitem__(self, key):
         del self._current_obj()[key]
+
+    def __call__(self, *args, **kw):
+        return self._current_obj()(*args, **kw)
     
     def __repr__(self):
         try:
             return repr(self._current_obj())
         except (TypeError, AttributeError):
-            return '<%s.%s object at 0x%08x>' % (__name__,
-                                                   self.__class__.__name__,
-                                                   id(self))
+            return '<%s.%s object at 0x%x>' % (self.__class__.__module__,
+                                               self.__class__.__name__,
+                                               id(self))
     
     def __iter__(self):
         return iter(self._current_obj())
@@ -161,14 +174,6 @@ class StackedObjectProxy(object):
     def __nonzero__(self):
         return bool(self._current_obj())
 
-    def current_obj(self):
-        """
-        Deprecated (Aug 15 2006); moved to _current_obj.
-        """
-        warnings.warn('StackedObjectProxy.current_obj has been moved to '
-                      'StackedObjectProxy._current_obj', DeprecationWarning, 2)
-        return self._current_obj()
-
     def _current_obj(self):
         """Returns the current active object being proxied to
         
@@ -176,7 +181,10 @@ class StackedObjectProxy(object):
         provided will be used. Otherwise, a TypeError will be raised.
         
         """
-        objects = getattr(self.____local__, 'objects', None)
+        try:
+            objects = self.____local__.objects
+        except AttributeError:
+            objects = None
         if objects:
             return objects[-1]
         else:
@@ -188,20 +196,12 @@ class StackedObjectProxy(object):
                     'No object (name: %s) has been registered for this '
                     'thread' % self.____name__)
 
-    def push_object(self, obj):
-        """
-        Deprecated (Aug 15 2006); moved to _push_object.
-        """
-        warnings.warn('StackedObjectProxy.push_object has been moved to '
-                      'StackedObjectProxy._push_object', DeprecationWarning, 2)
-        self._push_object(obj)
-
     def _push_object(self, obj):
         """Make ``obj`` the active object for this thread-local.
         
         This should be used like:
         
-        .. code-block:: Python
+        .. code-block:: python
 
             obj = yourobject()
             module.glob = StackedObjectProxy()
@@ -212,18 +212,12 @@ class StackedObjectProxy(object):
                 module.glob._pop_object(conf)
         
         """
-        if not hasattr(self.____local__, 'objects'):
+        try:
+            self.____local__.objects.append(obj)
+        except AttributeError:
             self.____local__.objects = []
-        self.____local__.objects.append(obj)
+            self.____local__.objects.append(obj)
     
-    def pop_object(self, obj=None):
-        """
-        Deprecated (Aug 15 2006); moved to _pop_object.
-        """
-        warnings.warn('StackedObjectProxy.pop_object has been moved to '
-                      'StackedObjectProxy._pop_object', DeprecationWarning, 2)
-        self._pop_object(obj)
-
     def _pop_object(self, obj=None):
         """Remove a thread-local object.
         
@@ -231,14 +225,24 @@ class StackedObjectProxy(object):
         error is emitted if they don't match.
         
         """
-        if not hasattr(self.____local__, 'objects'):
-            raise AssertionError('No object has been registered for this thread')
-        popped = self.____local__.objects.pop()
-        if obj:
-            if popped is not obj:
+        try:
+            popped = self.____local__.objects.pop()
+            if obj and popped is not obj:
                 raise AssertionError(
                     'The object popped (%s) is not the same as the object '
                     'expected (%s)' % (popped, obj))
+        except AttributeError:
+            raise AssertionError('No object has been registered for this thread')
+
+    def _object_stack(self):
+        """Returns all of the objects stacked in this container
+
+        (Might return [] if there are none)
+        """
+        try:
+            return self.____local__.objects[:]
+        except AssertionError:
+            return []
 
     # The following methods will be swapped for their original versions by
     # StackedObjectRestorer when restoration is enabled. The original
@@ -303,27 +307,41 @@ class Registry(object):
     
     def register(self, stacked, obj):
         """Register an object with a StackedObjectProxy"""
-        stacked._push_object(obj)
         myreglist = self.reglist[-1]
-        myreglist[id(stacked)] = (stacked, obj)
-
-    def replace(self, stacked, obj):
-        """Replace the object referenced by a StackedObjectProxy with a
-        different object
-
-        In the event that no object has been registered, the new object will
-        be registered.
+        stacked_id = id(stacked)
+        if stacked_id in myreglist:
+            stacked._pop_object(myreglist[stacked_id][1])
+            del myreglist[stacked_id]
+        stacked._push_object(obj)
+        myreglist[stacked_id] = (stacked, obj)
+    
+    def multiregister(self, stacklist):
+        """Register a list of tuples
+        
+        Similar call semantics as register, except this registers
+        multiple objects at once.
+        
+        Example::
+            
+            registry.multiregister([(sop, obj), (anothersop, anotherobj)])
+        
         """
         myreglist = self.reglist[-1]
-        if id(stacked) in myreglist:
-            stacked._pop_object(myreglist[id(stacked)][1])
-        self.register(stacked, obj)
+        for stacked, obj in stacklist:
+            stacked_id = id(stacked)
+            if stacked_id in myreglist:
+                stacked._pop_object(myreglist[stacked_id][1])
+                del myreglist[stacked_id]
+            stacked._push_object(obj)
+            myreglist[stacked_id] = (stacked, obj)
+    
+    # Replace now does the same thing as register
+    replace = register
     
     def cleanup(self):
         """Remove all objects from all StackedObjectProxy instances that
         were tracked at this Registry context"""
-        for id, val in self.reglist[-1].iteritems():
-            stacked, obj = val
+        for stacked, obj in self.reglist[-1].itervalues():
             stacked._pop_object(obj)
         self.reglist.pop()
         
@@ -342,13 +360,17 @@ class RegistryManager(object):
     object which is a Registry instance.
         
     """
-    def __init__(self, application):
+    def __init__(self, application, streaming=False):
         self.application = application
+        self.streaming = streaming
         
     def __call__(self, environ, start_response):
         app_iter = None
         reg = environ.setdefault('paste.registry', Registry())
         reg.prepare()
+        if self.streaming:
+            return self.streaming_iter(reg, environ, start_response)
+        
         try:
             app_iter = self.application(environ, start_response)
         except Exception, e:
@@ -377,6 +399,36 @@ class RegistryManager(object):
             reg.cleanup()
         
         return app_iter
+    
+    def streaming_iter(self, reg, environ, start_response):
+        try:
+            for item in self.application(environ, start_response):
+                yield item
+        except Exception, e:
+            # Regardless of if the content is an iterable, generator, list
+            # or tuple, we clean-up right now. If its an iterable/generator
+            # care should be used to ensure the generator has its own ref
+            # to the actual object
+            if environ.get('paste.evalexception'):
+                # EvalException is present in the WSGI stack
+                expected = False
+                for expect in environ.get('paste.expected_exceptions', []):
+                    if isinstance(e, expect):
+                        expected = True
+                if not expected:
+                    # An unexpected exception: save state for EvalException
+                    restorer.save_registry_state(environ)
+            reg.cleanup()
+            raise
+        except:
+            # Save state for EvalException if it's present
+            if environ.get('paste.evalexception'):
+                restorer.save_registry_state(environ)
+            reg.cleanup()
+            raise
+        else:
+            reg.cleanup()
+
 
 class StackedObjectRestorer(object):
     """Track StackedObjectProxies and their proxied objects for automatic
